
// Copyright (c) 2014 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#ifndef CSSPropertyNames_h
#define CSSPropertyNames_h

#include "core/CoreExport.h"
#include "wtf/Assertions.h"
#include <stddef.h>

namespace WTF {
class AtomicString;
class String;
}

namespace blink {

enum CSSPropertyID {
    CSSPropertyInvalid = 0,
    // This isn't a property, but we need to know the position of @apply rules in style rules
    CSSPropertyApplyAtRule = 1,
    CSSPropertyVariable = 2,
    CSSPropertyAnimationDelay = 3,
    CSSPropertyAnimationDirection = 4,
    CSSPropertyAnimationDuration = 5,
    CSSPropertyAnimationFillMode = 6,
    CSSPropertyAnimationIterationCount = 7,
    CSSPropertyAnimationName = 8,
    CSSPropertyAnimationPlayState = 9,
    CSSPropertyAnimationTimingFunction = 10,
    CSSPropertyTransitionDelay = 11,
    CSSPropertyTransitionDuration = 12,
    CSSPropertyTransitionProperty = 13,
    CSSPropertyTransitionTimingFunction = 14,
    CSSPropertyColor = 15,
    CSSPropertyDirection = 16,
    CSSPropertyFontFamily = 17,
    CSSPropertyFontKerning = 18,
    CSSPropertyFontSize = 19,
    CSSPropertyFontSizeAdjust = 20,
    CSSPropertyFontStretch = 21,
    CSSPropertyFontStyle = 22,
    CSSPropertyFontVariantLigatures = 23,
    CSSPropertyFontVariantCaps = 24,
    CSSPropertyFontVariantNumeric = 25,
    CSSPropertyFontWeight = 26,
    CSSPropertyFontFeatureSettings = 27,
    CSSPropertyWebkitFontSmoothing = 28,
    CSSPropertyWebkitLocale = 29,
    CSSPropertyTextOrientation = 30,
    CSSPropertyWebkitTextOrientation = 31,
    CSSPropertyWritingMode = 32,
    CSSPropertyWebkitWritingMode = 33,
    CSSPropertyTextRendering = 34,
    CSSPropertyZoom = 35,
    CSSPropertyAlignContent = 36,
    CSSPropertyAlignItems = 37,
    CSSPropertyAlignmentBaseline = 38,
    CSSPropertyAlignSelf = 39,
    CSSPropertyBackdropFilter = 40,
    CSSPropertyBackfaceVisibility = 41,
    CSSPropertyBackgroundAttachment = 42,
    CSSPropertyBackgroundBlendMode = 43,
    CSSPropertyBackgroundClip = 44,
    CSSPropertyBackgroundColor = 45,
    CSSPropertyBackgroundImage = 46,
    CSSPropertyBackgroundOrigin = 47,
    CSSPropertyBackgroundPositionX = 48,
    CSSPropertyBackgroundPositionY = 49,
    CSSPropertyBackgroundRepeatX = 50,
    CSSPropertyBackgroundRepeatY = 51,
    CSSPropertyBackgroundSize = 52,
    CSSPropertyBaselineShift = 53,
    CSSPropertyBorderBottomColor = 54,
    CSSPropertyBorderBottomLeftRadius = 55,
    CSSPropertyBorderBottomRightRadius = 56,
    CSSPropertyBorderBottomStyle = 57,
    CSSPropertyBorderBottomWidth = 58,
    CSSPropertyBorderCollapse = 59,
    CSSPropertyBorderImageOutset = 60,
    CSSPropertyBorderImageRepeat = 61,
    CSSPropertyBorderImageSlice = 62,
    CSSPropertyBorderImageSource = 63,
    CSSPropertyBorderImageWidth = 64,
    CSSPropertyBorderLeftColor = 65,
    CSSPropertyBorderLeftStyle = 66,
    CSSPropertyBorderLeftWidth = 67,
    CSSPropertyBorderRightColor = 68,
    CSSPropertyBorderRightStyle = 69,
    CSSPropertyBorderRightWidth = 70,
    CSSPropertyBorderTopColor = 71,
    CSSPropertyBorderTopLeftRadius = 72,
    CSSPropertyBorderTopRightRadius = 73,
    CSSPropertyBorderTopStyle = 74,
    CSSPropertyBorderTopWidth = 75,
    CSSPropertyBottom = 76,
    CSSPropertyBoxShadow = 77,
    CSSPropertyBoxSizing = 78,
    CSSPropertyBreakAfter = 79,
    CSSPropertyBreakBefore = 80,
    CSSPropertyBreakInside = 81,
    CSSPropertyBufferedRendering = 82,
    CSSPropertyCaptionSide = 83,
    CSSPropertyClear = 84,
    CSSPropertyClip = 85,
    CSSPropertyClipPath = 86,
    CSSPropertyClipRule = 87,
    CSSPropertyColorInterpolation = 88,
    CSSPropertyColorInterpolationFilters = 89,
    CSSPropertyColorRendering = 90,
    CSSPropertyColumnFill = 91,
    CSSPropertyContain = 92,
    CSSPropertyContent = 93,
    CSSPropertyCounterIncrement = 94,
    CSSPropertyCounterReset = 95,
    CSSPropertyCursor = 96,
    CSSPropertyCx = 97,
    CSSPropertyCy = 98,
    CSSPropertyD = 99,
    CSSPropertyDisplay = 100,
    CSSPropertyDominantBaseline = 101,
    CSSPropertyEmptyCells = 102,
    CSSPropertyFill = 103,
    CSSPropertyFillOpacity = 104,
    CSSPropertyFillRule = 105,
    CSSPropertyFilter = 106,
    CSSPropertyFlexBasis = 107,
    CSSPropertyFlexDirection = 108,
    CSSPropertyFlexGrow = 109,
    CSSPropertyFlexShrink = 110,
    CSSPropertyFlexWrap = 111,
    CSSPropertyFloat = 112,
    CSSPropertyFloodColor = 113,
    CSSPropertyFloodOpacity = 114,
    CSSPropertyGridAutoColumns = 115,
    CSSPropertyGridAutoFlow = 116,
    CSSPropertyGridAutoRows = 117,
    CSSPropertyGridColumnEnd = 118,
    CSSPropertyGridColumnGap = 119,
    CSSPropertyGridColumnStart = 120,
    CSSPropertyGridRowEnd = 121,
    CSSPropertyGridRowGap = 122,
    CSSPropertyGridRowStart = 123,
    CSSPropertyGridTemplateAreas = 124,
    CSSPropertyGridTemplateColumns = 125,
    CSSPropertyGridTemplateRows = 126,
    CSSPropertyHeight = 127,
    CSSPropertyHyphens = 128,
    CSSPropertyImageRendering = 129,
    CSSPropertyImageOrientation = 130,
    CSSPropertyIsolation = 131,
    CSSPropertyJustifyContent = 132,
    CSSPropertyJustifyItems = 133,
    CSSPropertyJustifySelf = 134,
    CSSPropertyLeft = 135,
    CSSPropertyLetterSpacing = 136,
    CSSPropertyLightingColor = 137,
    CSSPropertyLineHeight = 138,
    CSSPropertyListStyleImage = 139,
    CSSPropertyListStylePosition = 140,
    CSSPropertyListStyleType = 141,
    CSSPropertyMarginBottom = 142,
    CSSPropertyMarginLeft = 143,
    CSSPropertyMarginRight = 144,
    CSSPropertyMarginTop = 145,
    CSSPropertyMarkerEnd = 146,
    CSSPropertyMarkerMid = 147,
    CSSPropertyMarkerStart = 148,
    CSSPropertyMask = 149,
    CSSPropertyMaskSourceType = 150,
    CSSPropertyMaskType = 151,
    CSSPropertyMaxHeight = 152,
    CSSPropertyMaxWidth = 153,
    CSSPropertyMinHeight = 154,
    CSSPropertyMinWidth = 155,
    CSSPropertyMixBlendMode = 156,
    CSSPropertyObjectFit = 157,
    CSSPropertyObjectPosition = 158,
    CSSPropertyOffsetAnchor = 159,
    CSSPropertyOffsetDistance = 160,
    CSSPropertyOffsetPath = 161,
    CSSPropertyOffsetPosition = 162,
    CSSPropertyOffsetRotate = 163,
    CSSPropertyOffsetRotation = 164,
    CSSPropertyOpacity = 165,
    CSSPropertyOrder = 166,
    CSSPropertyOrphans = 167,
    CSSPropertyOutlineColor = 168,
    CSSPropertyOutlineOffset = 169,
    CSSPropertyOutlineStyle = 170,
    CSSPropertyOutlineWidth = 171,
    CSSPropertyOverflowAnchor = 172,
    CSSPropertyOverflowWrap = 173,
    CSSPropertyOverflowX = 174,
    CSSPropertyOverflowY = 175,
    CSSPropertyPaddingBottom = 176,
    CSSPropertyPaddingLeft = 177,
    CSSPropertyPaddingRight = 178,
    CSSPropertyPaddingTop = 179,
    CSSPropertyPaintOrder = 180,
    CSSPropertyPerspective = 181,
    CSSPropertyPerspectiveOrigin = 182,
    CSSPropertyPointerEvents = 183,
    CSSPropertyPosition = 184,
    CSSPropertyQuotes = 185,
    CSSPropertyResize = 186,
    CSSPropertyRight = 187,
    CSSPropertyR = 188,
    CSSPropertyRx = 189,
    CSSPropertyRy = 190,
    CSSPropertyScrollBehavior = 191,
    CSSPropertyScrollSnapType = 192,
    CSSPropertyScrollSnapPointsX = 193,
    CSSPropertyScrollSnapPointsY = 194,
    CSSPropertyScrollSnapDestination = 195,
    CSSPropertyScrollSnapCoordinate = 196,
    CSSPropertyShapeImageThreshold = 197,
    CSSPropertyShapeMargin = 198,
    CSSPropertyShapeOutside = 199,
    CSSPropertyShapeRendering = 200,
    CSSPropertySize = 201,
    CSSPropertySnapHeight = 202,
    CSSPropertySpeak = 203,
    CSSPropertyStopColor = 204,
    CSSPropertyStopOpacity = 205,
    CSSPropertyStroke = 206,
    CSSPropertyStrokeDasharray = 207,
    CSSPropertyStrokeDashoffset = 208,
    CSSPropertyStrokeLinecap = 209,
    CSSPropertyStrokeLinejoin = 210,
    CSSPropertyStrokeMiterlimit = 211,
    CSSPropertyStrokeOpacity = 212,
    CSSPropertyStrokeWidth = 213,
    CSSPropertyTableLayout = 214,
    CSSPropertyTabSize = 215,
    CSSPropertyTextAlign = 216,
    CSSPropertyTextAlignLast = 217,
    CSSPropertyTextAnchor = 218,
    CSSPropertyTextCombineUpright = 219,
    CSSPropertyTextDecoration = 220,
    CSSPropertyTextDecorationColor = 221,
    CSSPropertyTextDecorationLine = 222,
    CSSPropertyTextDecorationSkip = 223,
    CSSPropertyTextDecorationStyle = 224,
    CSSPropertyTextIndent = 225,
    CSSPropertyTextJustify = 226,
    CSSPropertyTextOverflow = 227,
    CSSPropertyTextShadow = 228,
    CSSPropertyTextSizeAdjust = 229,
    CSSPropertyTextTransform = 230,
    CSSPropertyTextUnderlinePosition = 231,
    CSSPropertyTop = 232,
    CSSPropertyTouchAction = 233,
    CSSPropertyTransform = 234,
    CSSPropertyTransformOrigin = 235,
    CSSPropertyTransformStyle = 236,
    CSSPropertyTranslate = 237,
    CSSPropertyRotate = 238,
    CSSPropertyScale = 239,
    CSSPropertyUnicodeBidi = 240,
    CSSPropertyVectorEffect = 241,
    CSSPropertyVerticalAlign = 242,
    CSSPropertyVisibility = 243,
    CSSPropertyX = 244,
    CSSPropertyY = 245,
    CSSPropertyWebkitAppearance = 246,
    CSSPropertyWebkitAppRegion = 247,
    CSSPropertyWebkitBackgroundClip = 248,
    CSSPropertyWebkitBackgroundOrigin = 249,
    CSSPropertyWebkitBorderHorizontalSpacing = 250,
    CSSPropertyWebkitBorderImage = 251,
    CSSPropertyWebkitBorderVerticalSpacing = 252,
    CSSPropertyWebkitBoxAlign = 253,
    CSSPropertyWebkitBoxDecorationBreak = 254,
    CSSPropertyWebkitBoxDirection = 255,
    CSSPropertyWebkitBoxFlex = 256,
    CSSPropertyWebkitBoxFlexGroup = 257,
    CSSPropertyWebkitBoxLines = 258,
    CSSPropertyWebkitBoxOrdinalGroup = 259,
    CSSPropertyWebkitBoxOrient = 260,
    CSSPropertyWebkitBoxPack = 261,
    CSSPropertyWebkitBoxReflect = 262,
    CSSPropertyColumnCount = 263,
    CSSPropertyColumnGap = 264,
    CSSPropertyColumnRuleColor = 265,
    CSSPropertyColumnRuleStyle = 266,
    CSSPropertyColumnRuleWidth = 267,
    CSSPropertyColumnSpan = 268,
    CSSPropertyColumnWidth = 269,
    CSSPropertyWebkitHighlight = 270,
    CSSPropertyWebkitHyphenateCharacter = 271,
    CSSPropertyWebkitLineBreak = 272,
    CSSPropertyWebkitLineClamp = 273,
    CSSPropertyWebkitMarginAfterCollapse = 274,
    CSSPropertyWebkitMarginBeforeCollapse = 275,
    CSSPropertyWebkitMarginBottomCollapse = 276,
    CSSPropertyWebkitMarginTopCollapse = 277,
    CSSPropertyWebkitMaskBoxImageOutset = 278,
    CSSPropertyWebkitMaskBoxImageRepeat = 279,
    CSSPropertyWebkitMaskBoxImageSlice = 280,
    CSSPropertyWebkitMaskBoxImageSource = 281,
    CSSPropertyWebkitMaskBoxImageWidth = 282,
    CSSPropertyWebkitMaskClip = 283,
    CSSPropertyWebkitMaskComposite = 284,
    CSSPropertyWebkitMaskImage = 285,
    CSSPropertyWebkitMaskOrigin = 286,
    CSSPropertyWebkitMaskPositionX = 287,
    CSSPropertyWebkitMaskPositionY = 288,
    CSSPropertyWebkitMaskRepeatX = 289,
    CSSPropertyWebkitMaskRepeatY = 290,
    CSSPropertyWebkitMaskSize = 291,
    CSSPropertyWebkitPerspectiveOriginX = 292,
    CSSPropertyWebkitPerspectiveOriginY = 293,
    CSSPropertyWebkitPrintColorAdjust = 294,
    CSSPropertyWebkitRtlOrdering = 295,
    CSSPropertyWebkitRubyPosition = 296,
    CSSPropertyWebkitTapHighlightColor = 297,
    CSSPropertyWebkitTextCombine = 298,
    CSSPropertyWebkitTextEmphasisColor = 299,
    CSSPropertyWebkitTextEmphasisPosition = 300,
    CSSPropertyWebkitTextEmphasisStyle = 301,
    CSSPropertyWebkitTextFillColor = 302,
    CSSPropertyWebkitTextSecurity = 303,
    CSSPropertyWebkitTextStrokeColor = 304,
    CSSPropertyWebkitTextStrokeWidth = 305,
    CSSPropertyWebkitTransformOriginX = 306,
    CSSPropertyWebkitTransformOriginY = 307,
    CSSPropertyWebkitTransformOriginZ = 308,
    CSSPropertyWebkitUserDrag = 309,
    CSSPropertyWebkitUserModify = 310,
    CSSPropertyUserSelect = 311,
    CSSPropertyWhiteSpace = 312,
    CSSPropertyWidows = 313,
    CSSPropertyWidth = 314,
    CSSPropertyWillChange = 315,
    CSSPropertyWordBreak = 316,
    CSSPropertyWordSpacing = 317,
    CSSPropertyWordWrap = 318,
    CSSPropertyZIndex = 319,
    CSSPropertyWebkitBorderEndColor = 320,
    CSSPropertyWebkitBorderEndStyle = 321,
    CSSPropertyWebkitBorderEndWidth = 322,
    CSSPropertyWebkitBorderStartColor = 323,
    CSSPropertyWebkitBorderStartStyle = 324,
    CSSPropertyWebkitBorderStartWidth = 325,
    CSSPropertyWebkitBorderBeforeColor = 326,
    CSSPropertyWebkitBorderBeforeStyle = 327,
    CSSPropertyWebkitBorderBeforeWidth = 328,
    CSSPropertyWebkitBorderAfterColor = 329,
    CSSPropertyWebkitBorderAfterStyle = 330,
    CSSPropertyWebkitBorderAfterWidth = 331,
    CSSPropertyWebkitMarginEnd = 332,
    CSSPropertyWebkitMarginStart = 333,
    CSSPropertyWebkitMarginBefore = 334,
    CSSPropertyWebkitMarginAfter = 335,
    CSSPropertyWebkitPaddingEnd = 336,
    CSSPropertyWebkitPaddingStart = 337,
    CSSPropertyWebkitPaddingBefore = 338,
    CSSPropertyWebkitPaddingAfter = 339,
    CSSPropertyWebkitLogicalWidth = 340,
    CSSPropertyWebkitLogicalHeight = 341,
    CSSPropertyWebkitMinLogicalWidth = 342,
    CSSPropertyWebkitMinLogicalHeight = 343,
    CSSPropertyWebkitMaxLogicalWidth = 344,
    CSSPropertyWebkitMaxLogicalHeight = 345,
    CSSPropertyAll = 346,
    CSSPropertyPage = 347,
    CSSPropertyWebkitFontSizeDelta = 348,
    CSSPropertyWebkitTextDecorationsInEffect = 349,
    CSSPropertyFontDisplay = 350,
    CSSPropertyMaxZoom = 351,
    CSSPropertyMinZoom = 352,
    CSSPropertyOrientation = 353,
    CSSPropertySrc = 354,
    CSSPropertyUnicodeRange = 355,
    CSSPropertyUserZoom = 356,
    CSSPropertyAnimation = 357,
    CSSPropertyBackground = 358,
    CSSPropertyBackgroundPosition = 359,
    CSSPropertyBackgroundRepeat = 360,
    CSSPropertyBorder = 361,
    CSSPropertyBorderBottom = 362,
    CSSPropertyBorderColor = 363,
    CSSPropertyBorderImage = 364,
    CSSPropertyBorderLeft = 365,
    CSSPropertyBorderRadius = 366,
    CSSPropertyBorderRight = 367,
    CSSPropertyBorderSpacing = 368,
    CSSPropertyBorderStyle = 369,
    CSSPropertyBorderTop = 370,
    CSSPropertyBorderWidth = 371,
    CSSPropertyFlex = 372,
    CSSPropertyFlexFlow = 373,
    CSSPropertyFont = 374,
    CSSPropertyFontVariant = 375,
    CSSPropertyGrid = 376,
    CSSPropertyGridArea = 377,
    CSSPropertyGridColumn = 378,
    CSSPropertyGridGap = 379,
    CSSPropertyGridRow = 380,
    CSSPropertyGridTemplate = 381,
    CSSPropertyListStyle = 382,
    CSSPropertyMargin = 383,
    CSSPropertyMarker = 384,
    CSSPropertyMotion = 385,
    CSSPropertyOffset = 386,
    CSSPropertyOutline = 387,
    CSSPropertyOverflow = 388,
    CSSPropertyPadding = 389,
    CSSPropertyPageBreakAfter = 390,
    CSSPropertyPageBreakBefore = 391,
    CSSPropertyPageBreakInside = 392,
    CSSPropertyTransition = 393,
    CSSPropertyWebkitBorderAfter = 394,
    CSSPropertyWebkitBorderBefore = 395,
    CSSPropertyWebkitBorderEnd = 396,
    CSSPropertyWebkitBorderStart = 397,
    CSSPropertyWebkitColumnBreakAfter = 398,
    CSSPropertyWebkitColumnBreakBefore = 399,
    CSSPropertyWebkitColumnBreakInside = 400,
    CSSPropertyColumnRule = 401,
    CSSPropertyColumns = 402,
    CSSPropertyWebkitMarginCollapse = 403,
    CSSPropertyWebkitMask = 404,
    CSSPropertyWebkitMaskBoxImage = 405,
    CSSPropertyWebkitMaskPosition = 406,
    CSSPropertyWebkitMaskRepeat = 407,
    CSSPropertyWebkitTextEmphasis = 408,
    CSSPropertyWebkitTextStroke = 409,
    CSSPropertyAliasMotionOffset = 672,
    CSSPropertyAliasMotionPath = 673,
    CSSPropertyAliasMotionRotation = 676,
    CSSPropertyAliasEpubCaptionSide = 595,
    CSSPropertyAliasEpubTextCombine = 810,
    CSSPropertyAliasEpubTextEmphasis = 920,
    CSSPropertyAliasEpubTextEmphasisColor = 811,
    CSSPropertyAliasEpubTextEmphasisStyle = 813,
    CSSPropertyAliasEpubTextOrientation = 543,
    CSSPropertyAliasEpubTextTransform = 742,
    CSSPropertyAliasEpubWordBreak = 828,
    CSSPropertyAliasEpubWritingMode = 545,
    CSSPropertyAliasWebkitAlignContent = 548,
    CSSPropertyAliasWebkitAlignItems = 549,
    CSSPropertyAliasWebkitAlignSelf = 551,
    CSSPropertyAliasWebkitAnimation = 869,
    CSSPropertyAliasWebkitAnimationDelay = 515,
    CSSPropertyAliasWebkitAnimationDirection = 516,
    CSSPropertyAliasWebkitAnimationDuration = 517,
    CSSPropertyAliasWebkitAnimationFillMode = 518,
    CSSPropertyAliasWebkitAnimationIterationCount = 519,
    CSSPropertyAliasWebkitAnimationName = 520,
    CSSPropertyAliasWebkitAnimationPlayState = 521,
    CSSPropertyAliasWebkitAnimationTimingFunction = 522,
    CSSPropertyAliasWebkitBackfaceVisibility = 553,
    CSSPropertyAliasWebkitBackgroundSize = 564,
    CSSPropertyAliasWebkitBorderBottomLeftRadius = 567,
    CSSPropertyAliasWebkitBorderBottomRightRadius = 568,
    CSSPropertyAliasWebkitBorderRadius = 878,
    CSSPropertyAliasWebkitBorderTopLeftRadius = 584,
    CSSPropertyAliasWebkitBorderTopRightRadius = 585,
    CSSPropertyAliasWebkitBoxShadow = 589,
    CSSPropertyAliasWebkitBoxSizing = 590,
    CSSPropertyAliasWebkitClipPath = 598,
    CSSPropertyAliasWebkitColumnCount = 775,
    CSSPropertyAliasWebkitColumnGap = 776,
    CSSPropertyAliasWebkitColumnRule = 913,
    CSSPropertyAliasWebkitColumnRuleColor = 777,
    CSSPropertyAliasWebkitColumnRuleStyle = 778,
    CSSPropertyAliasWebkitColumnRuleWidth = 779,
    CSSPropertyAliasWebkitColumnSpan = 780,
    CSSPropertyAliasWebkitColumnWidth = 781,
    CSSPropertyAliasWebkitColumns = 914,
    CSSPropertyAliasWebkitFilter = 618,
    CSSPropertyAliasWebkitFlex = 884,
    CSSPropertyAliasWebkitFlexBasis = 619,
    CSSPropertyAliasWebkitFlexDirection = 620,
    CSSPropertyAliasWebkitFlexFlow = 885,
    CSSPropertyAliasWebkitFlexGrow = 621,
    CSSPropertyAliasWebkitFlexShrink = 622,
    CSSPropertyAliasWebkitFlexWrap = 623,
    CSSPropertyAliasWebkitFontFeatureSettings = 539,
    CSSPropertyAliasWebkitJustifyContent = 644,
    CSSPropertyAliasWebkitOpacity = 677,
    CSSPropertyAliasWebkitOrder = 678,
    CSSPropertyAliasWebkitPerspective = 693,
    CSSPropertyAliasWebkitPerspectiveOrigin = 694,
    CSSPropertyAliasWebkitShapeImageThreshold = 709,
    CSSPropertyAliasWebkitShapeMargin = 710,
    CSSPropertyAliasWebkitShapeOutside = 711,
    CSSPropertyAliasWebkitTextSizeAdjust = 741,
    CSSPropertyAliasWebkitTransform = 746,
    CSSPropertyAliasWebkitTransformOrigin = 747,
    CSSPropertyAliasWebkitTransformStyle = 748,
    CSSPropertyAliasWebkitTransition = 905,
    CSSPropertyAliasWebkitTransitionDelay = 523,
    CSSPropertyAliasWebkitTransitionDuration = 524,
    CSSPropertyAliasWebkitTransitionProperty = 525,
    CSSPropertyAliasWebkitTransitionTimingFunction = 526,
    CSSPropertyAliasWebkitUserSelect = 823,
};

const int firstCSSProperty = 3;
const int numCSSProperties = 407;
const int lastCSSProperty = 409;
const int lastUnresolvedCSSProperty = 920;
const size_t maxCSSPropertyNameLength = 40;

const char* getPropertyName(CSSPropertyID);
const WTF::AtomicString& getPropertyNameAtomicString(CSSPropertyID);
WTF::String getPropertyNameString(CSSPropertyID);
WTF::String getJSPropertyName(CSSPropertyID);

inline bool isCSSPropertyIDWithName(int id)
{
    return id >= firstCSSProperty && id <= lastUnresolvedCSSProperty;
}

inline bool isValidCSSPropertyID(CSSPropertyID id)
{
    return id != CSSPropertyInvalid;
}

inline CSSPropertyID convertToCSSPropertyID(int value)
{
    DCHECK(value >= CSSPropertyInvalid && value <= lastCSSProperty);
    return static_cast<CSSPropertyID>(value);
}

inline CSSPropertyID resolveCSSPropertyID(CSSPropertyID id)
{
    return convertToCSSPropertyID(id & ~512);
}

inline bool isPropertyAlias(CSSPropertyID id) { return id & 512; }

CSSPropertyID unresolvedCSSPropertyID(const WTF::String&);

CSSPropertyID CORE_EXPORT cssPropertyID(const WTF::String&);

} // namespace blink

#endif // CSSPropertyNames_h
